from GMM_Demux import compute_venn
from math import pow
from math import log
from scipy.stats import binom
from scipy.stats import binomtest

def compute_multiplet_rates_asymp(cell_num, sample_num, drop_num):
    no_drop_rate = (1 - 1 / drop_num)
    cells_per_sample = round(cell_num / sample_num)
    drop_with_cells = (1 - pow(no_drop_rate, cell_num)) * drop_num
    single_sample_drops = sample_num * (1 - pow(no_drop_rate, cells_per_sample) ) * drop_num * pow(no_drop_rate, (sample_num - 1) * cells_per_sample)
    singlet_drops = cell_num * pow(no_drop_rate, cell_num - 1)

    singlet_rate = singlet_drops / drop_with_cells
    single_sample_rate = single_sample_drops / drop_with_cells
    MSM_rate = 1 - single_sample_rate
    SSM_rate = single_sample_rate - singlet_rate
    return MSM_rate, SSM_rate, singlet_rate, drop_with_cells


def compute_relative_SSM_rate_asymp(cell_num, drop_num):
    no_drop_rate = (1 - 1 / drop_num)
    drop_with_cells = (1 - pow(no_drop_rate, cell_num)) * drop_num
    singlet_drops = cell_num * pow(no_drop_rate, cell_num - 1)

    return 1 - singlet_drops / drop_with_cells


def compute_relative_SSM_rate(SSM_rate, singlet_rate):
    return SSM_rate / singlet_rate


def get_min_hto_num(cell_num, drop_num, SSM_threshold, sample_num = 1):
    sample_num = 1
    while True:
        #(MSM_rate, SSM_rate, singlet_rate) = compute_multiplet_rates(cell_num, sample_num, drop_num)
        (MSM_rate, SSM_rate, singlet_rate, drop_with_cells) = compute_multiplet_rates_asymp(cell_num, sample_num, drop_num)
        relative_SSM_rate = compute_relative_SSM_rate(SSM_rate, singlet_rate)
        #print(relative_SSM_rate)
        if (relative_SSM_rate <= SSM_threshold):
            break;
        else:
            sample_num += 1

    return sample_num


def cell_num_estimator(a_num, captured_drop_num, capture_rate):
    estimated_drop_num = captured_drop_num / capture_rate
    base = 1 - 1 / estimated_drop_num
    power = 1 - a_num / captured_drop_num
    #print("power:", power)
    #print("base:", base)
    cell_num = log(power, base)
    return cell_num


def drop_num_estimator(a_num, b_num, shared_num):
    drop_num = a_num * b_num / shared_num 
    return drop_num 


def compute_shared_num(drop_num, A_num, B_num):
    A_rate = compute_mix_rate(drop_num, B_num) 
    #print("A_rate: ", A_rate)
    shared_num = A_rate *  A_num
    return shared_num


# Computes the rate of drops that have certain cells in them.
def compute_mix_rate(drop_num, cell_num):
    no_drop_rate = (1 - 1 / drop_num)
    cell_in_rate = 1 - pow(no_drop_rate, cell_num)
    return cell_in_rate


def compute_SSM_rate_with_cell_num(cell_num, drop_num):
    no_drop_rate = (1 - 1 / drop_num)
    singlet_drops = cell_num * pow(no_drop_rate, cell_num - 1)
    drop_with_cells = (1 - pow(no_drop_rate, cell_num)) * drop_num
    SSM_rate = 1 - singlet_drops / drop_with_cells
    return SSM_rate


def compute_SSD_num(drop_num, subject_cell_num, total_cell_num, ambiguous_rate = 0):
    no_drop_rate = (1 - 1 / drop_num)
    non_subject_num = total_cell_num - subject_cell_num
    SSD_prob = (1 - pow(no_drop_rate, subject_cell_num)) \
            * pow(no_drop_rate, int(non_subject_num * (1 - ambiguous_rate)))

    return int(SSD_prob * drop_num)


def compute_GEM_prob(drop_num, cell_num):
    return 1 - binom.pmf(0, cell_num, 1 / drop_num)


def phony_cluster_MSM_rate(cell_num_ary, cell_type_num = 2):
    assert(cell_type_num >= 2)

    total_cell_num = sum(cell_num_ary)
    sample_prob_ary = [cell_num / total_cell_num for cell_num in cell_num_ary]
    
    return 1 - sum([pow(sample_prob, cell_type_num) for sample_prob in sample_prob_ary])


def get_tau_cell_num(drop_num, total_cell_num, cluster_GEM_num, ambiguous_rate = 0.0):
    SSD_prob = cluster_GEM_num / drop_num 
    no_drop_rate = 1 - 1 / drop_num
    certain_rate = 1 - ambiguous_rate
    
    tau_cell_num = total_cell_num - int(log(SSD_prob \
            + pow(no_drop_rate, total_cell_num * certain_rate + cluster_GEM_num * ambiguous_rate) \
            , no_drop_rate) / certain_rate)

    return tau_cell_num


def pure_cluster_MSM_rate(drop_num, cluster_GEM_num, cell_num_ary, capture_rate, ambiguous_rate = 0):
    #print("==================")

    total_cell_num = sum(cell_num_ary)
    sample_prob_ary = [cell_num / total_cell_num for cell_num in cell_num_ary]

    cluster_GEM_num = cluster_GEM_num / capture_rate

    #print(cluster_GEM_num) 

    mix_cell_num = get_tau_cell_num(drop_num, total_cell_num, cluster_GEM_num)
    tau_cell_num = mix_cell_num 
    #tau_cell_num = (mix_cell_num - total_cell_num * ambiguous_rate) / (1 - ambiguous_rate)
    #tau_cell_num = get_tau_cell_num(drop_num, total_cell_num, cluster_GEM_num)

    #print("initial tau_cell_num: ", tau_cell_num)

    pure_type_GEM_num = compute_SSD_num(drop_num, tau_cell_num + (total_cell_num - tau_cell_num) * ambiguous_rate, total_cell_num, ambiguous_rate)
    while (pure_type_GEM_num > cluster_GEM_num):
        tau_cell_num -= 1
        pure_type_GEM_num = compute_SSD_num(drop_num, tau_cell_num + (total_cell_num - tau_cell_num) * ambiguous_rate, total_cell_num, ambiguous_rate)

    #print("final tau_cell_num: ", tau_cell_num)

    tau_num_ary = [int((tau_cell_num + (total_cell_num - tau_cell_num) * ambiguous_rate) * sample_prob) for sample_prob in sample_prob_ary]
    SSD_num_ary = [compute_SSD_num(drop_num, tau_num, total_cell_num) for tau_num in tau_num_ary]
    total_SSD_num = sum(SSD_num_ary)

    #print("tau GEM num: ", pure_type_GEM_num)
    #print("total num: ", compute_SSD_num(drop_num, total_cell_num, total_cell_num))

    return 1 - (total_SSD_num / pure_type_GEM_num)


def test_phony_hypothesis(cluster_MSM_num, cluster_GEM_num, cell_num_ary, capture_rate):
    MSM_rate = phony_cluster_MSM_rate(cell_num_ary)
    return binomtest(cluster_MSM_num / capture_rate, cluster_GEM_num / capture_rate, MSM_rate, "less")


def test_pure_hypothesis(cluster_MSM_num, drop_num, cluster_GEM_num, cell_num_ary, capture_rate, ambiguous_rate = 0):
    MSM_rate = pure_cluster_MSM_rate(drop_num, cluster_GEM_num, cell_num_ary, capture_rate, ambiguous_rate)
    print("Estimated MSM rate: ", MSM_rate)
    return binomtest(cluster_MSM_num / capture_rate, cluster_GEM_num / capture_rate, MSM_rate, "greater")


####Debuging Functions####
def debug_get_cell_num(drop_num, GEM_num, capture_rate):
    no_drop_rate = (1 - 1 / drop_num)
    return log(1 - GEM_num / (drop_num * capture_rate), no_drop_rate)


def debug_compute_doublet_num(drop_num, type_a_num, type_b_num):
    no_drop_rate = (1 - 1 / drop_num)
    SSD_prob = (1 - pow(no_drop_rate, type_a_num)) * (1 - pow(no_drop_rate, type_b_num))

    return int(SSD_prob * drop_num)


def debug_pure_cluster_MSM_rate(drop_num, tau_cell_num, sample_num_ary, capture_rate, ambiguous_rate = 0):
    total_cell_num = sum(sample_num_ary)
    sample_prob_ary = [sample_num / total_cell_num for sample_num in sample_num_ary]

    tau_num_ary = [int((tau_cell_num + (total_cell_num - tau_cell_num) * ambiguous_rate) * sample_prob) for sample_prob in sample_prob_ary]

    print(tau_num_ary)

    SSD_num_ary = [compute_SSD_num(drop_num, tau_num, total_cell_num) for tau_num in tau_num_ary]

    print(SSD_num_ary)

    total_SSD_num = sum(SSD_num_ary)

    print(total_SSD_num)

    pure_type_GEM_num = compute_SSD_num(drop_num, tau_cell_num + (total_cell_num - tau_cell_num) * ambiguous_rate, total_cell_num)
    #pure_type_GEM_num = compute_SSD_num(drop_num, tau_cell_num + (total_cell_num - tau_cell_num) * ambiguous_rate, total_cell_num * (1 - ambiguous_rate))

    print("Pure type num: ", int(pure_type_GEM_num * capture_rate))

    return 1 - (total_SSD_num / pure_type_GEM_num)
####End of Debuging Functions####


def compute_observation_probability(drop_num, capture_rate, cell_num_ary, HTO_GEM_ary, base_bv_array, sample_num):
    log_probability = 0

    GEM_prob_ary = []

    #print(drop_num, capture_rate, cell_num_ary, HTO_GEM_ary)
    
    for sample_idx in range(sample_num):
        ori_GEM_num = round(HTO_GEM_ary[sample_idx] / capture_rate)
        GEM_formation_prob = compute_GEM_prob(drop_num, cell_num_ary[sample_idx])
        GEM_prob_ary.append(GEM_formation_prob)

    for i in range(len(HTO_GEM_ary)):
        bv_idx = i + 1
        GEM_formation_prob = 1

        for sample_idx in range(sample_num):
            if compute_venn.check_set_bit(base_bv_array[bv_idx], sample_idx, sample_num):
                GEM_formation_prob *= GEM_prob_ary[sample_idx]

        ori_GEM_num = round(HTO_GEM_ary[i] / capture_rate)
        sample_binom_prob = binom.pmf(ori_GEM_num, drop_num, GEM_formation_prob)
        #print("***ori_GEM_num:", ori_GEM_num)
        #print("***GEM_formation_prob:", GEM_formation_prob)
        #print("***sample_binom_prob:", sample_binom_prob)
        #print("***log sample_binom_prob:", log(sample_binom_prob))
        #print("***log sample_binom_prob, corrected:", log(sample_binom_prob) * ((1/sample_num) ** (base_bv_array[bv_idx].count_bits() - 1)))
        log_probability += log(sample_binom_prob) * ((1/sample_num) ** (base_bv_array[bv_idx].count_bits() - 1))
        #probability *= sample_binom_prob

    return log_probability



# Just to suppress sklearn import imp warning
def warn(*args, **kwargs):
    pass
import warnings
warnings.warn = warn

import pandas as pd
from GMM_Demux import compute_venn
from GMM_Demux import estimator
from GMM_Demux import classify_drops
from GMM_Demux import GMM_IO
from sys import argv
import sys
import argparse
from tabulate import tabulate
from scipy.special import comb

def main():
    ####### Parsing parameters and preparing data #######
    parser = argparse.ArgumentParser(prog='GMM-demux', conflict_handler='resolve')

    # Positional arguments have * number of arguments atm.
    parser.add_argument('input_path', help = "The input path of mtx files from cellRanger pipeline.", nargs="*")
    parser.add_argument('hto_array', help = "Names of the HTO tags, separated by ','.", nargs="*")

    # Optional arguments.
    parser.add_argument("-k", "--skip", help="Load a full classification report and skip the mtx folder. Requires a path argument to the full report folder. When specified, the user no longer needs to provide the mtx folder.", type=str)
    parser.add_argument("-x", "--extract", help="Names of the HTO tag(s) to extract, separated by ','. Joint HTO samples are combined with '+', such as 'HTO_1+HTO_2'.", type=str)
    parser.add_argument("-o", "--output", help="The path for storing the Same-Sample-Droplets (SSDs). SSDs are stored in mtx format. Requires a path argument.", type=str, default="SSD_mtx")
    parser.add_argument("-f", "--full", help="Generate the full classification report. Requires a path argument.", type=str)
    parser.add_argument("-c", "--csv", help="Take input in csv format, instead of mmx format.", action='store_true')
    parser.add_argument("-t", "--threshold", help="Provide the confidence threshold value. Requires a float in (0,1). Default value: 0.8", type=float, default=0.8)
    parser.add_argument("-s", "--simplified", help="Generate the simplified classification report. Requires a path argument.", type=str)
    parser.add_argument("-u", "--summary", help = "Generate the statstic summary of the dataset. Including MSM, SSM rates. Requires an estimated total number of cells in the assay as input.", type=int)
    parser.add_argument("-r", "--report", help="Store the data summary report. Requires a file argument. Only executes if -u is set.", type=str)
    parser.add_argument("-e", "--examine", help="Provide the cell list. Requires a file argument. Only executes if -u is set.", type=str)
    parser.add_argument("-a", "--ambiguous", help="The estimated chance of having a phony GEM getting included in a pure type GEM cluster by the clustering algorithm. Requires a float in (0, 1). Default value: 0.05. Only executes if -e executes.", type=float, default=0.05)
    parser.add_argument("-rs", "--random_seed", help = "If provided, this is passed to the GaussianMixture random_state parameter.", default=None, type=int)
    
    print("==============================GMM-Demux Initialization==============================")

    args = parser.parse_args()

    confidence_threshold = args.threshold
    print("Confidence threshold:", confidence_threshold)


    # Classify droplets
    if not args.skip:
        # Overwrite the positional arguments
        parser.add_argument('input_path', help = "The input path of mtx files from cellRanger pipeline.")
        parser.add_argument('hto_array', help = "Names of the HTO tags, separated by ','.")
        args = parser.parse_args()

        input_path = args.input_path
        hto_array = args.hto_array.split(',')

        output_path = args.output
        print("Output directory:", output_path)

        #TODO: add CLR to csv data.
        if args.csv:
            full_df, GMM_df = GMM_IO.read_csv(input_path, hto_array)
        else:
            full_df, GMM_df = GMM_IO.read_cellranger(input_path, hto_array)
        
        GEM_num = GMM_df.shape[0]
        sample_num = GMM_df.shape[1]


        ####### Run classifier #######
        base_bv_array = compute_venn.obtain_base_bv_array(sample_num)
        #print([int(i) for i in base_bv_array])
        (high_array, low_array) = classify_drops.obtain_arrays(GMM_df, args.random_seed)

        # Obtain extract array.
        if args.extract:
            extract_id_ary = []
            tag_name_ary = []

            for tag_name in args.extract.split(','):
                tag_name_ary.append(tag_name.split('+') )

            for tag_ary in tag_name_ary:
                mask = compute_venn.init_mask(sample_num)
                for tag in tag_ary:
                    hto_idx = hto_array.index(tag)
                    bv = compute_venn.set_bit(mask, hto_idx)

                for idx in range(0, len(base_bv_array) ):
                    if base_bv_array[idx] == mask:
                        extract_id = idx

                extract_id_ary.append(extract_id)

        else:
            extract_id_ary = None 


        # Obtain classification result
        GMM_full_df, class_name_ary = \
                classify_drops.classify_drops(base_bv_array, high_array, low_array, sample_num, GEM_num, GMM_df.index, GMM_df.columns.values)

        # Store classification results
        if args.full:
            print("Full classification result is stored in", args.full)
            classify_drops.store_full_classify_result(GMM_full_df, class_name_ary, args.full)

        if args.simplified:
            ########## Paper Specific ############
            #purified_df = classify_drops.purify_droplets(GMM_full_df, confidence_threshold)
            ########## Paper Specific ############
            print("Simplified classification result is stored in", args.simplified)
            classify_drops.store_simplified_classify_result(GMM_full_df, class_name_ary, args.simplified, sample_num, confidence_threshold)
        
        # Clean up bad drops
        purified_df = classify_drops.purify_droplets(GMM_full_df, confidence_threshold)

        # Store SSD result
        print("MSM-free droplets are stored in folder", output_path, "\n")
        
        SSD_idx = classify_drops.obtain_SSD_list(purified_df, sample_num, extract_id_ary)
        SSD_df = GMM_IO.store_cellranger(full_df, SSD_idx, output_path)

        # Record sample names for summary report.
        sampe_names = GMM_df.columns

    # Parse the full report.
    else:
        GMM_full_df, sample_num, class_name_ary, sampe_names = classify_drops.read_full_classify_result(args.skip)
        base_bv_array = compute_venn.obtain_base_bv_array(sample_num)
        purified_df = classify_drops.purify_droplets(GMM_full_df, confidence_threshold)
        SSD_idx = classify_drops.obtain_SSD_list(purified_df, sample_num)


    ####### If extract is eanbled, other functions are disabled #######
    if args.extract:
        exit()


    ####### Estimate SSM #######
    if args.summary:
        # Count bad drops
        negative_num, unclear_num = classify_drops.count_bad_droplets(GMM_full_df, confidence_threshold)

        estimated_total_cell_num = args.summary

        # Infer parameters
        HTO_GEM_ary = compute_venn.obtain_HTO_GEM_num(purified_df, base_bv_array, sample_num)

        params0 = [80000, 0.5]

        for i in range(sample_num):
            params0.append(round(HTO_GEM_ary[i] * estimated_total_cell_num / sum(HTO_GEM_ary[:sample_num])))

        combination_counter = 0
        try:
            for i in range(1, sample_num + 1):
                combination_counter += comb(sample_num, i, True)
                HTO_GEM_ary_main = HTO_GEM_ary[0:combination_counter]
                params0 = compute_venn.obtain_experiment_params(base_bv_array, HTO_GEM_ary_main, sample_num, estimated_total_cell_num, params0)
        except:
            print("GMM cannot find a viable solution that satisfies the droplet formation model. SSM rate estimation terminated.")
            sys.exit(0)
                

        # Legacy parameter estimation
        #(cell_num_ary, drop_num, capture_rate) = compute_venn.obtain_HTO_cell_n_drop_num(purified_df, base_bv_array, sample_num, estimated_total_cell_num, confidence_threshold)
        (drop_num, capture_rate, *cell_num_ary) = params0

        SSM_rate_ary = [estimator.compute_SSM_rate_with_cell_num(cell_num_ary[i], drop_num) for i in range(sample_num)]
        rounded_cell_num_ary = [round(cell_num) for cell_num in cell_num_ary]
        SSD_count_ary = classify_drops.get_SSD_count_ary(purified_df, SSD_idx, sample_num)
        count_ary = classify_drops.count_by_class(purified_df, base_bv_array)
        MSM_rate, SSM_rate, singlet_rate = compute_venn.gather_multiplet_rates(count_ary, SSM_rate_ary, sample_num)

        # Generate report
        full_report_dict = {
            "#Drops": round(drop_num),
            "Capture rate": "%5.2f" % (capture_rate * 100),
            "#Cells": sum(rounded_cell_num_ary),
            "Singlet": "%5.2f" % (singlet_rate * 100),
            "MSM": "%5.2f" % (MSM_rate * 100),
            "SSM": "%5.2f" % (SSM_rate * 100),
            "RSSM": "%5.2f" % (estimator.compute_relative_SSM_rate(SSM_rate, singlet_rate) * 100),
            "Negative": "%5.2f" % (negative_num / GMM_full_df.shape[0] * 100),
            "Unclear": "%5.2f" % (unclear_num / GMM_full_df.shape[0] * 100)
            }
        full_report_columns = [
            "#Drops",
            "Capture rate",
            "#Cells",
            "Singlet",
            "MSM",
            "SSM",
            "RSSM",
            "Negative",
            "Unclear"
            ]

        full_report_df = pd.DataFrame(full_report_dict, index = ["Total"], columns=full_report_columns)

        print("==============================Full Report==============================")
        print(tabulate(full_report_df, headers='keys', tablefmt='psql'))
        print ("\n\n")
        print("==============================Per Sample Report==============================")
        sample_df = pd.DataFrame(data=[
            ["%d" % num for num in rounded_cell_num_ary],
            ["%d" % num for num in SSD_count_ary],
            ["%5.2f" % (num * 100) for num in SSM_rate_ary]
            ],
            columns = sampe_names, index = ["#Cells", "#SSDs", "RSSM"])
        print(tabulate(sample_df, headers='keys', tablefmt='psql'))

        if args.report:
            print("\n\n***Summary report is stored in folder", args.report)
            with open(args.report, "w") as report_file:
                report_file.write("==============================Full Report==============================\n")
            with open(args.report, "a") as report_file:
                report_file.write(tabulate(full_report_df, headers='keys', tablefmt='psql'))
            with open(args.report, "a") as report_file:
                report_file.write("\n\n")
                report_file.write("==============================Per Sample Report==============================\n")
            with open(args.report, "a") as report_file:
                report_file.write(tabulate(sample_df, headers='keys', tablefmt='psql'))


        # Verify cell type 
        if args.examine:
            print("\n\n==============================Verifying the GEM Cluster==============================")

            ambiguous_rate = args.ambiguous
            print("Ambiguous rate:", ambiguous_rate)

            simplified_df = classify_drops.store_simplified_classify_result(purified_df, class_name_ary, None, sample_num, confidence_threshold)

            cell_list_path = args.examine
            cell_list = [line.rstrip('\n') for line in open(args.examine)]
            cell_list = list(set(cell_list).intersection(simplified_df.index.tolist()))

            ########## Paper Specific ############
            #cell_list_df = pd.read_csv(args.examine, index_col = 0)
            #cell_list = cell_list_df.index.tolist()
            ########## Paper Specific ############

            MSM_list = classify_drops.obtain_MSM_list(simplified_df, sample_num, cell_list)

            GEM_num = len(cell_list)
            MSM_num = len(MSM_list)
            print("GEM count: ", GEM_num, " | MSM count: ", MSM_num)

            phony_test_pvalue = estimator.test_phony_hypothesis(MSM_num, GEM_num, rounded_cell_num_ary, capture_rate)
            pure_test_pvalue = estimator.test_pure_hypothesis(MSM_num, drop_num, GEM_num, rounded_cell_num_ary, capture_rate, ambiguous_rate)

            print("Phony-type testing. P-value: ", phony_test_pvalue)
            print("Pure-type testing. P-value: ", pure_test_pvalue)
            
            cluster_type = ""

            if phony_test_pvalue < 0.01 and pure_test_pvalue > 0.01:
                cluster_type = " pure"
            elif pure_test_pvalue < 0.01 and phony_test_pvalue > 0.01:
                cluster_type = " phony"
            else:
                cluster_type = "n unclear"

            print("Conclusion: The cluster is a" + cluster_type + " cluster.")

            ########## Paper Specific ############
            #estimated_phony_cluster_MSM_rate = estimator.phony_cluster_MSM_rate(rounded_cell_num_ary, cell_type_num = 2)
            #estimated_pure_cluster_MSM_rate = estimator.pure_cluster_MSM_rate(drop_num, GEM_num, rounded_cell_num_ary, capture_rate, ambiguous_rate)
            #print(str(estimated_phony_cluster_MSM_rate)+","+str(estimated_pure_cluster_MSM_rate)+","+str(MSM_num / float(GEM_num))+","+str(phony_test_pvalue)+","+str(pure_test_pvalue)+","+str(GEM_num / float(purified_df.shape[0]))+","+str(cluster_type), file=sys.stderr)
            ########## Paper Specific ############


if __name__ == "__main__":
    main()



from GMM_Demux import estimator
from GMM_Demux import classify_drops
import BitVector
import numpy as np

from sys import argv
import pandas as pd
from statistics import mean
from scipy.optimize import minimize
from scipy import optimize 
from scipy.optimize import Bounds
from scipy.optimize import LinearConstraint

# Returns the binary array representing all combinations of cells
def obtain_base_bv_array(sample_num):
    base_bv_array = []
    bv_array_segments = [0]

    for i in range(sample_num):
        base_bv_array.append(BitVector.BitVector(size = sample_num) )
        base_bv_array[-1][i] = 1

    bv_array_segments.append(len(base_bv_array))

    for i in range(1, sample_num):
        for j in range(bv_array_segments[-2], bv_array_segments[-1]):
            first_zero = sample_num - base_bv_array[j].reverse().next_set_bit()

            for k in range(first_zero, sample_num):
                base_bv_array.append(base_bv_array[j] | base_bv_array[k])
        bv_array_segments.append(len(base_bv_array))

    return [BitVector.BitVector(intVal = 0, size = sample_num)] + base_bv_array


def get_empty_bv(sample_num):
    return BitVector.BitVector(size = sample_num)


# Return true if the bit_pos th bit of bv is 1
def check_set_bit(bv, bit_pos, sample_num):
    mask = BitVector.BitVector(size = sample_num)
    mask[bit_pos] = 1
    
    return int(mask & bv) != 0


def init_mask(sample_num):
    mask = BitVector.BitVector(size = sample_num)
    return mask


def set_bit(bv, bit_pos):
    bv[bit_pos] = 1
    return bv


def gather_multiplet_rates(venn_values, SSM_rate_ary, sample_num):
    total_drops = 0
    total_singlets = 0
    total_MSMs = 0
    total_SSMs = 0

    for i in range(len(venn_values)):
        if i < sample_num:
            singlet_num = venn_values[i] * (1 - SSM_rate_ary[i])
            SSM_num = venn_values[i] - singlet_num

            total_SSMs += SSM_num
            total_singlets += singlet_num

        else:
            total_MSMs += venn_values[i]

        total_drops += venn_values[i]

    singlet_rate = total_singlets / total_drops
    SSM_rate = total_SSMs / total_drops
    MSM_rate = total_MSMs / total_drops

    return MSM_rate, SSM_rate, singlet_rate


def obtain_HTO_GEM_num(data_df, base_bv_array, sample_num):
    HTO_GEM_ary = []

    # Obtain hto numbers
    for i in range(1, len(base_bv_array)):
        #print(base_bv_array[i])
        GEM_num = 0

        for j in range(len(base_bv_array)):
            if (base_bv_array[i] & base_bv_array[j] == base_bv_array[i]):
                GEM_num += (data_df["Cluster_id"] == j).sum()

        HTO_GEM_ary.append(GEM_num)
    
    #print(HTO_GEM_ary)
    return HTO_GEM_ary


def experiment_params_wrapper(params, HTO_GEM_ary, sample_num, scaler, base_bv_array, operator):
    #print("***Params: ", params)
    scaled_params = params.copy()

    for i in range(len(scaler)):
        scaled_params[i] = operator(params[i], scaler[i])

    drop_num = round(scaled_params[0])
    capture_rate = scaled_params[1]
    cell_num_ary = [round(cell_num) for cell_num in scaled_params[2:]]

    return -estimator.compute_observation_probability(drop_num, capture_rate, cell_num_ary, HTO_GEM_ary, base_bv_array, sample_num)


def param_scaling(params, scaler, operator):
    for i in range(len(scaler)):
        params[i] = operator(params[i], scaler[i])

    return params


def compute_scaler(params):
    scaler = []

    # Bigger steps for the first 2 params
    scaler.append(800 / params[0])
    scaler.append(500 / params[1])

    sample_share = []
    for param in params[2:]:
        sample_share.append( param / sum(params[2:]) * 40000)

    for share, param in zip(sample_share, params[2:]):
        scaler.append(share / param)

    return scaler


def obtain_experiment_params(base_bv_array, HTO_GEM_ary, sample_num, estimated_total_cell_num, params0 = None):
    drop_num = 80000
    capture_rate = 0.5
    cell_num_ary = [estimated_total_cell_num / sample_num for i in range(sample_num)]

    if params0 is None:
        params0 = [drop_num, capture_rate, *cell_num_ary]

    lower_bound = [1, 0.0] + [0 for i in range(sample_num)]
    upper_bound = [np.inf, 1.0] + [np.inf for i in range(sample_num)]

    # Compute scaler
    scaler = compute_scaler(params0)

    #print("params before scaling:", params0)
    # Scale all values
    params0 = param_scaling(params0, scaler, lambda x, y: x * y)
    lower_bound = param_scaling(lower_bound, scaler, lambda x, y: x * y)
    upper_bound = param_scaling(upper_bound, scaler, lambda x, y: x * y)

    #print("params:", params0)
    #print("bounds:", lower_bound, upper_bound)

    bounds = Bounds(lower_bound, upper_bound)

    # Linear constraints are not scaled in current version.
    constraint_func = [
            {"type": "ineq", "fun": lambda x: sum([a/b for a,b in zip(x[2:], scaler[2:])]) - estimated_total_cell_num * 0.99},
            {"type": "ineq", "fun": lambda x: - sum([a/b for a,b in zip(x[2:], scaler[2:])]) + estimated_total_cell_num * 1.01}
            ]

    res = minimize(experiment_params_wrapper, params0, args=(HTO_GEM_ary, sample_num, scaler, base_bv_array, lambda x,y: x/y), method='SLSQP', constraints=constraint_func, bounds=bounds, options={'verbose': 1, 'eps': 10, 'ftol' : 0.0000001})

    #print(res)
    final_param = param_scaling(res.x, scaler, lambda x, y: x / y)

    #print(final_param)
    return final_param 


def obtain_HTO_cell_n_drop_num(data_df, base_bv_array, sample_num, estimated_total_cell_num, confidence_threshold):
    hto_num_ary = []
    HTO_num_ary = []
    estimated_drop_num_ary = []

    # Obtain hto numbers
    for i in range(sample_num):
        hto_num = 0

        for j in range(len(base_bv_array)):
            if check_set_bit(base_bv_array[j], i, sample_num):
                hto_num += (data_df["Cluster_id"] == j).sum()

        hto_num_ary.append(hto_num)

    for hto_a_idx in range(sample_num):
        for hto_b_idx in range(hto_a_idx + 1, sample_num):

            shared_num = 0

            for j in range(len(base_bv_array)):
                if check_set_bit(base_bv_array[j], hto_a_idx, sample_num) and check_set_bit(base_bv_array[j], hto_b_idx, sample_num):
                    shared_num += (data_df["Cluster_id"] == j).sum()

            if shared_num > 0:
                estimated_drop_num = estimator.drop_num_estimator(hto_num_ary[hto_a_idx], hto_num_ary[hto_b_idx], shared_num)
                estimated_drop_num_ary.append(estimated_drop_num)


    captured_drop_num = mean(estimated_drop_num_ary)
    #print("Drop num ary:", estimated_drop_num_ary)
    #print("Total drop num:", drop_num)

    diff_num = 10000000000

    #for capture_rate in np.arange(1.0, 0.0, -0.005):
        #print("test: ", capture_rate)

    for capture_rate in np.arange(1.0, 0.0, -0.005):

        HTO_num_ary = []

        for hto_idx in range(sample_num):
            hto_num = hto_num_ary[hto_idx]
            cell_num = estimator.cell_num_estimator(hto_num, captured_drop_num, capture_rate)
            HTO_num_ary.append(cell_num)

        total_cell_num = sum(HTO_num_ary)
        #print("total_cell_num: ", total_cell_num)
        #print("abs(total_cell_num - estimated_total_cell_num): ", abs(total_cell_num - estimated_total_cell_num))

        if (abs(total_cell_num - estimated_total_cell_num) > diff_num):
            break

        diff_num = abs(total_cell_num - estimated_total_cell_num)


    #print(HTO_num_ary)
    return HTO_num_ary, captured_drop_num / capture_rate, capture_rate




import pandas as pd
from sys import argv
import numpy as np

def get_HTO_cell_idx(high_array, threshold):
    high_idx = np.argwhere(high_array > threshold)
    return high_idx


def compute_confidence(high_array, low_array, high_ary_idx, all_ary_idx):
    product = np.full(len(high_array[0]), 1.0)

    for idx in all_ary_idx:
        if idx in high_ary_idx:
            product = product * high_array[idx]
        else:
            product = product * low_array[idx]

    return product


def get_shared_cell_idx(high_array, low_array, high_ary_idx, all_ary_idx, threshold):
    product = compute_confidence(high_array, low_array, high_ary_idx, all_ary_idx)
    residual_idx = np.argwhere(product > threshold)
    return residual_idx


def get_HTO_cell_num(high_array, threshold):
    high_idx = get_HTO_cell_idx(high_array, threshold) 
    return len(high_idx)


def get_shared_cell_num(high_array, low_array, high_ary_idx, all_ary_idx, threshold):
    residual_idx = get_shared_cell_idx(high_array, low_array, high_ary_idx, all_ary_idx, threshold) 
    return len(residual_idx)

